/*
  StreamStar GPI tally for ATMEGA328 || Arduino UNO & compatible
  Copyright (c) 2018-2019 Cristian Secară https://www.secarica.ro
  Published under the GPLv3 license -- see https://www.gnu.org/licenses/gpl-3.0.en.html
  This code is used by the SSGPITally hardware interface presented at https://www.secarica.ro/index.php/technical/ssgpitally
*/

#include <Wire.h>
#include <string.h>

// begin user defined constants -->
String APP_DOMAIN = "StreamStar Tally";
String APP_VERSION = "1.6";

// valid values are 4, 6 or 8
int MAX_CAM = 8;

// if CONFIG_BY_HARDWARE = true => CONSIDER_A, CONSIDER_B and CONSIDER_CHECKSUM flags are based on Arduino pins 14, 15, and 16 (ATMEGA328 pins 23, 24 and 25) status
// if CONFIG_BY_HARDWARE = false => CONSIDER_A, CONSIDER_B and CONSIDER_CHECKSUM flags are set according to the boolean definitions next below
boolean CONFIG_BY_HARDWARE = false;

// these booleans are taken into account only if the above CONFIG_BY_HARDWARE = false
// A stands for Autocut
boolean CONSIDER_A = false;
// S stands for Camera Switch (Select)
boolean CONSIDER_S = true;
// undocumented (experimentally determined)
boolean CONSIDER_CHECKSUM = true;

// Arduino pins for GPI OUT (ATMEGA328 pins 13, 12, 11, 6, 5, 4, 15 and 14)
int BOARD_OUT_PINS[] = {7, 6, 5, 4, 3, 2, 9, 8};
// Arduino pins for GPI IN (ATMEGA328 pins 23, 24 and 25)
// pin 14 (ATMEGA328 pin 23) corresponds to CONSIDER_A
// pin 15 (ATMEGA328 pin 24) corresponds to CONSIDER_S
// pin 16 (ATMEGA328 pin 25) corresponds to CONSIDER_CHECKSUM
int BOARD_INP_PINS[] = {14, 15, 16};

// actual code is customized for LCD2S-162 interface from Modtronix (Serial LCD 16x2 Daughter Board) over i2c
boolean LCD_INSTALLED = false;
int LCD_COL_SIZE = 16;
int LCD_ROW_SIZE = 2;
// i2c address for LCD display interface
int LCD_I2C = 0x28;

// should be false in normal operation
boolean LCD_JUSTFORFUN = false;
// <-- end user defined constants

// undocumented (experimentally determined)
unsigned int checksum_byte_al[] = {0x57, 0x97, 0xd7, 0x16, 0x56, 0x96, 0xd6, 0x17};
unsigned int checksum_byte_ao[] = {0x37, 0xf7, 0xb7, 0x76, 0x36, 0xf6, 0xb6, 0x77};
unsigned int checksum_byte_sl[] = {0x1e, 0x5e, 0x9e, 0xde, 0x1f, 0x5f, 0x9f, 0xdf};
unsigned int checksum_byte_so[] = {0x7e, 0x3e, 0xfe, 0xbe, 0x7f, 0x3f, 0xff, 0xbf};

// various
unsigned int camNr;
unsigned int bytesCount;
unsigned int actualBit;
unsigned char inputByte;
boolean inputStringComplete;
boolean tallyStatus;
boolean considerMode;
boolean checksumMode;
String tallyType = "";

void setup() {
    // initialize digital pins
    int a;
    // outputs to relays
    for (a = 0; a < MAX_CAM; a ++) {
        pinMode(BOARD_OUT_PINS[a], OUTPUT);
    }
    // hardware config by means of pin status
    // ***** requires pull-up resistors *****
    for (a = 0; a < 3; a ++) {
        pinMode(BOARD_INP_PINS[a], INPUT);
    }

    // join bus as master
    Wire.begin();
    if (LCD_INSTALLED) {

        // create digits 1 to 8 on negative background
        // for cam 1
        Wire.beginTransmission(LCD_I2C);
        Wire.write(0x92);
        Wire.write(0x00);
        Wire.write(B11011);
        Wire.write(B10011);
        Wire.write(B11011);
        Wire.write(B11011);
        Wire.write(B11011);
        Wire.write(B11011);
        Wire.write(B10001);
        Wire.write(B11111);
        Wire.endTransmission(false);

        // for cam 2
        Wire.beginTransmission(LCD_I2C);
        Wire.write(0x92);
        Wire.write(0x01);
        Wire.write(B10001);
        Wire.write(B01110);
        Wire.write(B11110);
        Wire.write(B11101);
        Wire.write(B11011);
        Wire.write(B10111);
        Wire.write(B00000);
        Wire.write(B11111);
        Wire.endTransmission(false);

        // for cam 3
        Wire.beginTransmission(LCD_I2C);
        Wire.write(0x92);
        Wire.write(0x02);
        Wire.write(B00000);
        Wire.write(B11101);
        Wire.write(B11011);
        Wire.write(B11101);
        Wire.write(B11110);
        Wire.write(B01110);
        Wire.write(B10001);
        Wire.write(B11111);
        Wire.endTransmission(false);

        // for cam 4
        Wire.beginTransmission(LCD_I2C);
        Wire.write(0x92);
        Wire.write(0x03);
        Wire.write(B11101);
        Wire.write(B11001);
        Wire.write(B10101);
        Wire.write(B01101);
        Wire.write(B00000);
        Wire.write(B11101);
        Wire.write(B11101);
        Wire.write(B11111);
        Wire.endTransmission(false);

        // for cam 5
        Wire.beginTransmission(LCD_I2C);
        Wire.write(0x92);
        Wire.write(0x04);
        Wire.write(B00000);
        Wire.write(B01111);
        Wire.write(B00001);
        Wire.write(B11110);
        Wire.write(B11110);
        Wire.write(B01110);
        Wire.write(B10001);
        Wire.write(B11111);
        Wire.endTransmission(false);

        // for cam 6
        Wire.beginTransmission(LCD_I2C);
        Wire.write(0x92);
        Wire.write(0x05);
        Wire.write(B11001);
        Wire.write(B10111);
        Wire.write(B01111);
        Wire.write(B00001);
        Wire.write(B01110);
        Wire.write(B01110);
        Wire.write(B10001);
        Wire.write(B11111);
        Wire.endTransmission(false);

        // for cam 7
        Wire.beginTransmission(LCD_I2C);
        Wire.write(0x92);
        Wire.write(0x06);
        Wire.write(B00000);
        Wire.write(B11110);
        Wire.write(B11101);
        Wire.write(B11011);
        Wire.write(B10111);
        Wire.write(B10111);
        Wire.write(B10111);
        Wire.write(B11111);
        Wire.endTransmission(false);

        // for cam 8
        Wire.beginTransmission(LCD_I2C);
        Wire.write(0x92);
        Wire.write(0x07);
        Wire.write(B10001);
        Wire.write(B01110);
        Wire.write(B01110);
        Wire.write(B10001);
        Wire.write(B01110);
        Wire.write(B01110);
        Wire.write(B10001);
        Wire.write(B11111);
        Wire.endTransmission(false);

        Wire.beginTransmission(LCD_I2C);
        Wire.write(0x80);  // initiate display commands
        Wire.write(0x0c);  // clear display and go to first line
        Wire.write(APP_DOMAIN.c_str());
        // Wire.write(0x0a);  // go to beginning of next line
        Wire.write("v");
        Wire.write(APP_VERSION.c_str());
        Wire.write(" started");
        Wire.endTransmission(false);

        delay (3000);
    }

    if (LCD_INSTALLED && !LCD_JUSTFORFUN) {
        printCamToLcd();
    }

    // initialize serial
    Serial.begin(9600);

    // other init
    inputStringComplete = false;
    bytesCount= 0;
    actualBit = 0;
}

void loop() {
    // inputStringComplete = false at initial setup
    // actualBit = 0 at initial setup
    if (inputStringComplete) {
        int b;
        int c;
        for (b = 0; b < MAX_CAM; b ++) {
            c = (1 << b);
            if (b == camNr) {
                break;
            }
        }

        if (CONFIG_BY_HARDWARE) {
            considerMode = digitalRead(BOARD_INP_PINS[0]);
        }
        else {
            considerMode = CONSIDER_A;
        }

        if ((tallyType == "A") && considerMode) {
            if (tallyStatus) {
                digitalWrite(BOARD_OUT_PINS[camNr], HIGH);
                actualBit = actualBit | c;
            }
            else {
                digitalWrite(BOARD_OUT_PINS[camNr], LOW);
                actualBit = actualBit & (~c);
            }
            if (LCD_INSTALLED && !LCD_JUSTFORFUN) {
                printCamToLcd();
            }
        }

        if (CONFIG_BY_HARDWARE) {
            considerMode = digitalRead(BOARD_INP_PINS[1]);
        }
        else {
            considerMode = CONSIDER_S;
        }

        if ((tallyType == "S") && considerMode) {
            if (tallyStatus) {
                digitalWrite(BOARD_OUT_PINS[camNr], HIGH);
                actualBit = actualBit | c;
            }
            else {
                digitalWrite(BOARD_OUT_PINS[camNr], LOW);
                actualBit = actualBit & (~c);
            }
            if (LCD_INSTALLED && !LCD_JUSTFORFUN) {
                printCamToLcd();
            }
        }
        // start over
        inputStringComplete = false;
    }
}

void serialEvent() {
    // bytesCount = 0 at initial setup
    while (Serial.available()) {
        inputByte = (char)Serial.read();

        // start byte
        if (bytesCount == 0 && inputByte == 0x02) {
            bytesCount ++;
        }

        // switching type byte
        else if (bytesCount == 1) {
            if (inputByte == 0x41) {
                tallyType = "A";
                bytesCount ++;
            }
            else if (inputByte == 0x53) {
                tallyType = "S";
                bytesCount ++;
            }
            else {
                bytesCount = 0;
            }
        }

        // camera byte
        else if (bytesCount == 2) {
            boolean tallyType_valid = false;
            int d;
            for (d = 0; d < MAX_CAM; d ++) {
                if ((tallyType == "A" && inputByte == (0x31 + d)) || (tallyType == "S" && inputByte == (0x30 + d))) {
                    camNr = (0 + d);
                    bytesCount ++;
                    tallyType_valid = true;
                    break;
                }
            }
            if (tallyType_valid == false) {
                bytesCount = 0;
            }
        }

        // tally byte
        else if (bytesCount == 3) {
            if (inputByte == 0x4c) {
                tallyStatus = true;
                bytesCount ++;
            }
            else if (inputByte == 0x4f) {
                tallyStatus = false;
                bytesCount ++;
            }
            else {
                bytesCount = 0;
            }
        }

        // end byte
        else if (bytesCount == 4 && inputByte == 0x03) {
            bytesCount ++;
        }

        // sort of checksum byte (undocumented)
        else if (bytesCount == 5) {

            if (CONFIG_BY_HARDWARE) {
                checksumMode = digitalRead(BOARD_INP_PINS[2]);
            }
            else {
                checksumMode = CONSIDER_CHECKSUM;
            }

            if (checksumMode) {
                if (tallyType == "A" && tallyStatus && inputByte == checksum_byte_al[camNr]) {
                    inputStringComplete = true;
                }
                else if (tallyType == "A" && !tallyStatus  && inputByte == checksum_byte_ao[camNr]) {
                    inputStringComplete = true;
                }
                else if (tallyType == "S" && tallyStatus && inputByte == checksum_byte_sl[camNr]) {
                    inputStringComplete = true;
                }
                else if (tallyType == "S" && !tallyStatus && inputByte == checksum_byte_so[camNr]) {
                    inputStringComplete = true;
                }
            }
            else {
                inputStringComplete = true;
            }
            bytesCount ++;

            // just for fun
            if (LCD_INSTALLED && LCD_JUSTFORFUN && inputStringComplete) {
                printInputDataToLCD();
            }

        }

        // ready to get new serial data
        if (bytesCount > 5) {
            bytesCount = 0;
        }
    }
}

void printCamToLcd() {
    Wire.beginTransmission(LCD_I2C);
    Wire.write(0x80);  // initiate display commands
    Wire.write(0x0c);  // clear display and go to first line
    Wire.write("inverse = live");
    Wire.write(0x0a);  // go to beginning of next line
    Wire.write("camera ");
    int b;
    int c;
    for (b = 0; b < MAX_CAM; b ++) {
        c = (1 << b);
        if (actualBit & c) {
            Wire.write(0x00 + b);
        }
        else {
            Wire.write(0x31 + b);
        }
    }
    Wire.endTransmission(false);
}

// of no real use in normal operation
void printInputDataToLCD() {
    Wire.beginTransmission(LCD_I2C);
    Wire.write(0x80);
    Wire.write(0x0c);
    Wire.write("cam");
    Wire.write(camNr + 0x31);
    Wire.write(" typ");
    if (tallyType == "A") {
        Wire.write("A");
    }
    else if (tallyType == "S") {
        Wire.write("S");
    }
    else {
        Wire.write("?");
    }
    Wire.write(" tly");
    if (tallyStatus) {
        Wire.write("1");
    }
    else {
        Wire.write("0");
    }
    Wire.write(0x0a);
    Wire.write("checksum:0x");
    unsigned char inputByte_msb;
    inputByte_msb = int(inputByte) & 0xF0;
    inputByte_msb = inputByte_msb >> 4;
    Wire.write("0123456789abcdef"[inputByte_msb]);
    unsigned char inputByte_lsb;
    inputByte_lsb = int(inputByte) & 0x0F;
    Wire.write("0123456789abcdef"[inputByte_lsb]);
    Wire.endTransmission(false);
}
